/* -*-c-*-
 *
 * $Id: rabin.c,v 1.1 1999/11/19 13:17:57 mdw Exp $
 *
 * Miller-Rabin primality test
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: rabin.c,v $
 * Revision 1.1  1999/11/19 13:17:57  mdw
 * Prime number generator and tester.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include "mp.h"
#include "mpmont.h"
#include "pgen.h"
#include "rabin.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @rabin_create@ --- *
 *
 * Arguments:	@rabin *r@ = pointer to Rabin-Miller context
 *		@mp *m@ = pointer to number to test
 *
 * Returns:	---
 *
 * Use:		Precomputes some useful values for performing the
 *		Miller-Rabin probabilistic primality test.
 */

void rabin_create(rabin *r, mp *m)
{
  mp *m1 = mp_sub(MP_NEW, m, MP_ONE);
  mpscan sc;
  size_t s;

  /* --- Find @r@ and @s@ --- */

  mpmont_create(&r->mm, m);
  mp_scan(&sc, m1);
  s = 0;
  while (mp_step(&sc)) {
    if (mp_bit(&sc))
      break;
    s++;
  }
  r->s = s;
  r->r = mp_lsr(MP_NEW, m1, s);

  /* --- Compute %$(m - 1)R \bmod m$% --- */

  r->m1 = mpmont_mul(&r->mm, MP_NEW, m1, r->mm.r2);
  mp_drop(m1);
}

/* --- @rabin_destroy@ --- *
 *
 * Arguments:	@rabin *r@ = pointer to Rabin-Miller context
 *
 * Returns:	---
 *
 * Use:		Disposes of a Rabin-Miller context when it's no longer
 *		needed.
 */

void rabin_destroy(rabin *r)
{
  mp_drop(r->r);
  mp_drop(r->m1);
  mpmont_destroy(&r->mm);
}

/* --- @rabin_test@ --- *
 *
 * Arguments:	@rabin *r@ = pointer to Rabin-Miller context
 *		@mp *g@ = base to test the number against
 *
 * Returns:	Either @PGEN_COMPOSITE@ if the test failed, or @PGEN_MAYBE@
 *		if it succeeded.
 *
 * Use:		Performs a single iteration of the Rabin-Miller primality
 *		test.
 */

int rabin_test(rabin *r, mp *g)
{
  mp *y;
  mp *dd, *spare = MP_NEW;
  size_t j;
  int rc = PGEN_COMPOSITE;

  /* --- Calculate %$y R = g^r R \bmod m$% --- *
   *
   * If %$y = 1$% or %$y = m - 1$% then %$m$% is prime.  If course, note that
   * @y@ here has an extra factor of %$R$%.
   */

  y = mpmont_expr(&r->mm, g, r->r);
  if (MP_CMP(y, ==, r->mm.r) || MP_CMP(y, ==, r->m1)) {
    rc = PGEN_MAYBE;
    goto done;
  }

  /* --- Now for the main loop --- *
   *
   * If %$y^{2^j} \ne m - 1$% for any %$0 \le j < s$% then %$m$% is
   * composite.  Of course, %$j = 0$% has already been tested.
   */

  for (j = 1; j < r->s; j++) {
    dd = mpmont_mul(&r->mm, spare, y, y); spare = y; y = dd;
    if (MP_CMP(y, ==, r->mm.r))
      break;
    if (MP_CMP(y, ==, r->m1)) {
      rc = PGEN_MAYBE;
      break;
    }
  }

  /* --- Done --- */

done:
  if (spare != MP_NEW)
    MP_DROP(spare);
  MP_DROP(y);
  return (rc);
}

/*----- That's all, folks -------------------------------------------------*/
